function [E,partition] = calculate_eigenvalues(problem,pmin,pmax,tol,indices)
%CALCULATE_EIGENVALUES
%[E,partition] = calculate_eigenvalues(problem,pmin,pmax,tol,indices)

if ~isfield(problem,'partition') || isempty(problem.partition) || isinf(problem.b) || isinf(problem.a)
  %construct an initial mesh  
  partition=calculate_partition_init(problem,pmin,pmax,tol,indices);
else
  %an initial mesh was computed before, and passed here as input argument   
  partition=problem.partition;  
end
%compute eigenvalues
if indices
     %compute eigenvalues with index between pmin and pmax
     [E,partition] = calculate_eigenvalues_indices(partition,problem.bcs,pmin,pmax,tol);
else
    %compute eigenvalues between pmin and pmax
     [E,partition] = calculate_eigenvalues_interval(partition,problem.bcs,pmin,pmax,tol);
end
end

function [E,partition]= calculate_eigenvalues_indices(partition,bcs,nmin,nmax,tol)
%computes eigenvalues with index between nmin and nmax
%partition: contains information on the mesh and the associated data
%bcs: boundary conditions
%tol: user input tolerance
global messages

pfun=partition.pfun; qfun=partition.qfun; wfun=partition.wfun;
bp=partition.bp; rp=partition.rp;
bp.bcs=bcs; rp.bcs=bcs;
singular=false;
if bp.inda>1 || bp.indb>1
    singular=true;
end
E.eigenvalues = zeros(1,nmax-nmin+1);
E.errors= zeros(1,nmax-nmin+1);
E.indices= zeros(1,nmax-nmin+1);
E.status= ones(1,nmax-nmin+1);

%compute e_low such that E_{nmin-1}<e_low<E_nmin:
if messages
    disp('Computing lower bound of the requested eigenvalue')
end
[e_low,h_en,prufer]=get_lowerbound(nmin);
number_eigv=0;
while true
   	number_eigv = number_eigv + 1; %counts number of eigenvalues computed
    [n_low,phi_low]=locate_eigenvalue(e_low,rp);
    if prufer && n_low>nmax || number_eigv>nmax-nmin+1
        if ~prufer
            E.status=2*ones(1,number_eigv-1);
        end
        break;
    end
    if messages
        disp(' ');
        disp(['Computing eigenvalue ' num2str(number_eigv+nmin-1) ' ...'])
    end
    it=0;
    %find a bracket interval [e1, e2] for the eigenvalue:
    while true
      it=it+1;  
      e_up = e_low + h_en;
      [n_up,phi_up] = locate_eigenvalue(e_up,rp);
      if n_up<=n_low && phi_up*phi_low>0 
         e_low = e_up; h_en = h_en*2 ;
         n_low=n_up; 
      else
         while n_up>n_low+1 || isnan(n_up)
             e=(e_low+e_up)/2;
             n_t=locate_eigenvalue(e,rp);
             if n_t>n_low || isnan(n_t)
                 e_up=e;
                 n_up=n_t;
             else
                 e_low=e;
                 n_low=n_t;
             end
         end
  	     break; % stop while	
      end
    end %while
    e1=e_low;  e2=e_up;
 
    %compute eigenvalue approximation over the basic mesh (bp):
    [E1,Enear] = calculate_eigenvalues_secant(bp,e_low,e_up,tol);
    %compute the index:
    n_1 = locate_eigenvalue(E1,rp,true);
    if prufer && abs(nmin+number_eigv-1-n_1)>0.2  %close eigenvalues
           [E1,e_low,n_low,e_up]=calculate_eigenvalue_lin_interpolation(e1,e2,nmin+number_eigv-1,bp,tol);
     end
    %compute eigenvalue approximation over the reference mesh:
    if isinf(E1)
        E1=e_low;
        Enear=e_up;
    end
    E2 = calculate_eigenvalues_secant(rp,E1,Enear,tol/10); %E1 is used as initial estimate
    n_1 = locate_eigenvalue(E2,rp,true);
    if prufer && abs(nmin+number_eigv-1-n_1)>0.5
         [E2,e_low,n_low,e_up,n_up,n_1]=calculate_eigenvalue_lin_interpolation(e_low,e_up,nmin+number_eigv-1,rp,tol/10);
    end
    extrap=(64*E2-E1)/63;   
    if singular && abs(extrap-E1)>max(1,abs(E2)/100)*tol
          %(implicit) truncation point for the singular endpoint needs to be
          % closer to the singular endpoint
          E1t=E1; it=0;
          while abs(extrap-E1)>max(1,abs(E2)/100)*tol
            bps=bp;  it=it+1;
            if messages && it>1 
              disp('refining mesh near singular endpoint...')
              disp(['Number of mesh points: ' num2str(length(bp.xs))])
            end
            bp=calculate_partition_refine_sing(bp,pfun,qfun,wfun);
            bpt=bp;
            E1=E1t;
            E1t=calculate_eigenvalues_secant(bp,E1,Enear,tol);
            extrap=E1t;
          end
         bp=bps;
         if it>1
           %rp needs to be finer than bp:  
           rp=calculate_partition_refine(bpt,pfun,qfun,wfun);
           E1=E1t;   
           E2=calculate_eigenvalues_secant(rp,E1,Enear,tol);
           extrap=E2;
         else
            extrap=(64*E2-E1)/63; 
         end
    end
    Et=E1; Ett=Enear;
        while abs(extrap-E1)>max(1,abs(E2))*tol
          % mesh needs to be refined:  the number of meshpoints is doubled
          if singular
             bp=calculate_partition_refine(bp,pfun,qfun,wfun);
          else
              bp=rp;
          end
          rp=calculate_partition_refine(rp,pfun,qfun,wfun);
          if messages
                  disp('refining mesh...')
                  disp(['Number of mesh points: ' num2str(length(bp.xs))])
          end
          E1=E2; %use estimate over basic partition as initial estimate
          E2=calculate_eigenvalues_secant(rp,Et,Ett,tol/10);
          n_1 = locate_eigenvalue(E2,rp,true);
          if prufer && abs(nmin+number_eigv-1-n_1)>0.5
            [E2,Et,n_low,Ett,n_up,n_1]=calculate_eigenvalue_lin_interpolation(e_low,e_up,nmin+number_eigv-1,rp,tol/10);
          end
          extrap=E2;
        end
     
    E.eigenvalues(number_eigv) = E1;
    if abs(extrap-E1)<eps
      extrap=calculate_eigenvalues_secant(rp,E1,Enear,1e-12);
    end
    E.errors(number_eigv)= abs(extrap-E1);
    E.indices(number_eigv) = round(n_1);
    E.status(number_eigv)= round(n_1)~=nmin+number_eigv-1; %0 = OK
    h_en=min(h_en,abs(e_up-e_low));
    e_low = max(e_up,E1+tol);
    if messages
        disp(sprintf('Approximation of eigenvalue %d : %e \n',nmin+number_eigv-1,E1)); 
        disp('')
    end
 end %for
 partition.bp=bp; partition.rp=rp;
 
 
function [e_low,h_en,prufer]=get_lowerbound(nmin)
   %compute e_low such that E_{nmin-1}<E-low<E_{nmin} 
   %Based on the fact that lambda_k is approximately the root of 
   % int(sqrt(lambda*w-q)/p)=(k+1)*pi
   %where the integral is taken over the parts of the interval where
   %(lambda*w-q)/p >0
     
   
   rq0=rp.q0;  rw0=rp.w0;  rp0=rp.p0; 
   inda=rp.inda; indb=rp.indb;
   xs=rp.xs;
   hs=xs(rp.inf_a+2:end-rp.inf_b)-xs(rp.inf_a+1:end-1-rp.inf_b);
   if rp.inf_a %infinite endpoint a
     hs=[-1/xs(2) hs];
   end
   if rp.inf_b
     hs=[hs 1/xs(end-1)];
   end
   indices=inda+2:length(rq0)-indb-1;
   emin = min((rq0(indices)./rw0(indices)));
   emax = max(rq0(indices)./rw0(indices));
 
   
   if emax-emin>1e4
       emax=median((rq0(indices))./(rw0(indices)));
   end
   if emax==emin
       emax=emin+1;
   end
   tmp=(emin*rw0-rq0).*rp0;
   tmp(tmp<0)=0;
   index_emin=sum(sqrt(tmp).*hs)/pi-1;
   tmp=(emax*rw0-rq0).*rp0;
   tmp(tmp<0)=0;
   index_emax=sum(sqrt(tmp).*hs)/pi-1;   
   while index_emax<nmin %make sure emax is larger than E_{nmin}
         h_en = (emax-emin)/max((index_emax-index_emin),1);
         emax = emax+h_en;
         tmp=(emax*rw0-rq0).*rp0;
         tmp(tmp<0)=0;
         index_emax=sum(sqrt(tmp).*hs)/pi-1;
   end
   if index_emin>nmin
      if inda>0 && rp0(1)>1e15 && emin<-1e5
          prufer=false;
          index_emin=round(index_emin);
          index_e=index_emin;
          while index_e==index_emin
            emint=emin;
            emin=emin+(emax-emin)/10000;
            tmp=(emin*rw0-rq0).*rp0;
            tmp(tmp<0)=0;
            index_emin=round(sum(sqrt(tmp).*hs)/pi-1);
          end
          e_low=emint;
          h_en=1;
          return;
      end
   end
   while round(index_emin) >nmin  %make sure emin is smaller than E_{nmin}
        h_en = (emax-emin)/(index_emax-index_emin);
        emin=emin-h_en;
        tmp=(emin*rw0-rq0).*rp0;
        tmp(tmp<0)=0;
        index_emin=sum(sqrt(tmp).*hs)/pi-1;
   end

   while true
       e=(emax+emin)/2;
       tmp=(e*rw0-rq0).*rp0;
       tmp(tmp<0)=0;
       k=sum(sqrt(tmp).*hs)/pi-1;
       if k < nmin - 5 || (nmin<6 && k < nmin-0.5)
           emin=e;
           index_emin=k;
       elseif k>nmin+5
           emax=e;
           index_emax=k;
       else
           break;
       end   
   end
   h_en = (emax-emin)/(index_emax-index_emin)/10;
   if h_en==0
    h_en=1;
   end   
         
   e_low=emin;
   n_low_bp = locate_eigenvalue(e_low,bp);
   n_low_rp = locate_eigenvalue(e_low,rp);
   
   it=0;
   while (isnan(n_low_bp)  || isnan(n_low_rp)) && it<20
     e_low=e_low+h_en;
     n_low_bp = locate_eigenvalue(e_low,bp);
     n_low_rp = locate_eigenvalue(e_low,rp);
     it=it+1;
   end
  while abs(n_low_bp-n_low_rp)>0
      bp=rp; 
      if messages 
              disp('refining mesh...')
              disp(['Number of mesh points: ' num2str(length(bp.xs))])
      end
      n_low_bp=n_low_rp;
      rp=calculate_partition_refine_s(rp,pfun,qfun,wfun);
      n_low_rp = locate_eigenvalue(e_low,rp);
  end
      
 if locate_eigenvalue(emax,rp)>nmin
   while true   
     e_low1=(e_low+emax)/2;
     n_low1 = locate_eigenvalue(e_low1,rp);
     if n_low1 > nmin
       emax=e_low1;
     else
       e_low=e_low1;
       if locate_eigenvalue(e_low1,bp)~=n_low1
           bp=rp; 
           if messages 
              disp('refining mesh...')
              disp(['Number of mesh points: ' num2str(length(bp.xs))])
           end
           rp=calculate_partition_refine_s(rp,pfun,qfun,wfun);
       end
     end
     if n_low1 == nmin || abs(emax-e_low)<abs(emax)*eps
       e_low=e_low1; break;
     end
   end
   h_en=min(h_en,emax-e_low);
 else
   e_low=emax;
 end
  if h_en==0
    h_en=1;
  end 
   prufer=true;

end
end

function [E,partition]= calculate_eigenvalues_interval(partition,bcs,emin,emax,tol)
%compute eigenvalues between emin and emax
global messages
pfun=partition.pfun;qfun=partition.qfun;wfun=partition.wfun;
bp=partition.bp; rp=partition.rp;
bp.bcs=bcs; rp.bcs=bcs;
singular=false;
if bp.inda>1 || bp.indb>1
    singular=true;
end

number_eigv = 0; 
h_en = (emax-emin)/100; 
n_low_rp = locate_eigenvalue(emin,rp);
n_low_bp = locate_eigenvalue(emin,bp);
while n_low_rp~=n_low_bp
       bp=rp; 
      if messages 
              disp('refining mesh...')
              disp(['Number of mesh points: ' num2str(length(bp.xs))])
      end
      n_low_bp=n_low_rp;
      rp=calculate_partition_refine(rp,pfun,qfun,wfun);
      n_low_rp = locate_eigenvalue(emin,rp);
end
nmin=floor(n_low_rp);
e_low=emin;

 while e_low<emax
    [n_low,phi_low] = locate_eigenvalue(e_low,rp);
	number_eigv = number_eigv + 1;
    it=0;
    %find a bracket interval [e1, e2] for the eigenvalue:
    while true
      it=it+1;  
      e_up = e_low + h_en;
      [n_up,phi_up] = locate_eigenvalue(e_up,rp);
      if n_up<=n_low && phi_up*phi_low>0
         e_low = e_up;  h_en = h_en*2 ;
         n_low=n_up;  phi_low=phi_up;
         if e_low>emax
          return;
         end
      else
         while n_up>n_low+1
             e=(e_low+e_up)/2;
             n_t=locate_eigenvalue(e,rp); 
             if n_t>n_low
                 e_up=e;
                 n_up=n_t;
             else
                 e_low=e;
                 n_low=n_t;
             end
         end
	     break; % stop while	
      end
    end %while
    e1=e_low;   e2=e_up;
    %compute eigenvalue approximation over the basic mesh
    [E1,Enear] = calculate_eigenvalues_secant(bp,e_low,e_up,tol);
    %compute the index
    n_1 = locate_eigenvalue(E1,rp,true);
    if abs(nmin+number_eigv-1-n_1)>0.2  %close eigenvalues
          [E1,e_low,n_low,e_up]=calculate_eigenvalue_lin_interpolation(e1,e2,nmin+number_eigv-1,bp,tol);
    end
    %compute eigenvalue approximation over the reference mesh:
    E2 = calculate_eigenvalues_secant(rp,E1,Enear,tol/10);
    n_1 = locate_eigenvalue(E2,rp,true);
    if abs(nmin+number_eigv-1-n_1)>0.5
        [E2,e_low,n_low,e_up,n_up,n_1]=calculate_eigenvalue_lin_interpolation(e_low,e_up,nmin+number_eigv-1,rp,tol/10);
    end
    extrap=(64*E2-E1)/63;
    if singular && abs(extrap-E1)>max(1,abs(E2)/100)*tol
          %(implicit) truncation point for the singular endpoint needs to be
          % closer to the singular endpoint
          E1t=E1; it=0;
          while abs(extrap-E1)>max(1,abs(E2)/100)*tol
            bps=bp;  it=it+1;
            if messages && it>1 
              disp('refining mesh near singular endpoint...')
              disp(['Number of mesh points: ' num2str(length(bp.xs))])
            end
            bp=calculate_partition_refine_sing(bp,pfun,qfun,wfun);
            bpt=bp;
            E1=E1t;
            E1t=calculate_eigenvalues_secant(bp,E1,Enear,tol);
            extrap=E1t;
          end
         bp=bps;
         if it>1
           %rp needs to be finer than bp:  
           rp=calculate_partition_refine(bpt,pfun,qfun,wfun);
           E1=E1t;   
           E2=calculate_eigenvalues_secant(rp,E1,Enear,tol);
           extrap=E2;
         else
            extrap=(64*E2-E1)/63; 
         end
    end
    while abs(extrap-E1)>max(1,abs(E2))*tol
      % mesh needs to be refined:  the number of meshpoints is doubled   
      if singular
         bp=calculate_partition_refine(bp,pfun,qfun,wfun);
      else
          bp=rp;
      end
      rp=calculate_partition_refine(rp,pfun,qfun,wfun);
      if messages 
              disp('refining mesh...')
              disp(['Number of mesh points: ' num2str(length(bp.xs))])
      end
      E1=E2; %use estimate over basic partition as initial estimate
      E2=calculate_eigenvalues_secant(rp,E1,Enear,tol/10);
      n_1 = locate_eigenvalue(E2,rp,true);
      if  abs(nmin+number_eigv-1-n_1)>0.5
        [E2,e_low,n_low,e_up,n_up,n_1]=calculate_eigenvalue_lin_interpolation(e_low,e_up,nmin+number_eigv-1,rp,tol/10);
      end
      extrap=E2;
    end
    if E1>emax
          return;
    end
    E.eigenvalues(number_eigv) = E1;
    if abs(extrap-E1)<eps
      extrap=calculate_eigenvalues_secant(rp,E1,Enear,1e-12);
    end
    E.errors(number_eigv)= extrap-E1;
    E.indices(number_eigv) = round(n_1);
    E.status(number_eigv)= round(n_1)~=nmin+number_eigv-1;
    h_en=min(h_en,abs(e_up-e_low)/2);
    e_low = max(e_up,E1+tol);
    if messages
        disp(sprintf('Approximation of eigenvalue %d : %e \n',nmin+number_eigv-1,E1)); 
        disp('')
    end
 end %for
 partition.bp=bp; partition.rp=rp;
end

 

 
 function [e,e_low,n_low,e_up,n_up,pruf]=calculate_eigenvalue_lin_interpolation(e_low,e_up,n,part,tol0)
%   For given index n and lower and upper limits e_low and e_up, this 
%   function calculates the eigenvalue e_n (argument e) by using 
%   the Prufer phase function. Since this function uses linear 
%   interpolation, it is more robust but slower than calculate_eigenvalue_secant(). 
%   For these reasons calculate_eigenvalue_lin_interpolation() is called only 
%   when there are some hints that the results produced by 
%   calculate_eigenvalue_secant() may be inaccurate. 
% 
n_low = locate_eigenvalue(e_low,part,true);
n_up = locate_eigenvalue(e_up,part,true);

h_en=(e_up-e_low)/10;
while n_low>n
   e_low=e_low-h_en;
   n_low = locate_eigenvalue(e_low,part,true);
   h_en=h_en*2;
end
h_en=(e_up-e_low)/10;
while n_up<n
   e_up=e_up+h_en;
   n_up = locate_eigenvalue(e_up,part,true);
   h_en=h_en*2;
end

p_low = n_low-n;
p_up = n_up-n;
it=0;	
while true
    it = it + 1;
    if rem(it,2) == 0
        e = (e_low+e_up)/2;
    else
        e = (p_up*e_low-p_low*e_up)/(p_up-p_low); %linear interpolation    
    end
    pruf = locate_eigenvalue(e,part,true);
	pc = pruf-n;
	scalev = max(1,abs(e)/10); 
	tol = scalev*tol0;
    if abs(e_up-e_low) > tol %#1 %tol
        if p_low*pc < 0 %#2
         p_up = pc;
	     e_up = e;
         n_up = pruf;
	    else %#2
          if abs(p_low)<eps
            e = e_low;
            pruf = n_low;
            return;
          end
          if pc == 0
            return;
          end
  	        e_low = e;
            n_low = pruf;
	        p_low = pc;
        end %#2
	else %#1
	  break;
    end %#1
end %while
pruf=round(pruf);
 end

 
function [e1,e0] = calculate_eigenvalues_secant(part,e1,e2,tol)
%secans method to find root of mismatch function

%make sure e1 and e2 are close enough to eigenvalue
phi1=compute_mismatch(e1,part);
phi2=compute_mismatch(e2,part) ;
phi0=abs(phi1);
phit=phi0+1;
while abs(phit)>abs(phi0)
  et=e1-phi1*(e1-e2)/(phi1-phi2);
  phit=compute_mismatch(et,part);
  if phit*phi2>0
   e0=e2; phi0=phi2;
   e2=et;
   phi2=phit;
  else
    e0=e1; phi0=phi1;
    e1=et;
    phi1=phit;
  end  
end
it=1;

%secant method applied on mismatch function to locate eigenvalue
while abs(e1-e2)>max(abs(e1)*tol/10,1e-14)
 it=it+1;   
 et=e1;    
 e1=e1-phi1*(e1-e2)/(phi1-phi2);
 phi2=phi1;
 phi1=compute_mismatch(e1,part);    
 e0=e2;
 e2=et;
end
end



function phi=compute_mismatch(E,part)
%computes the mismatch function

if part.indb==0 % not singular in b
  imatch=length(part.p0);
  if part.inda>0 % singular in a
    if E*part.w0(1)-part.q0(1)<0
        y=[0;1];
    elseif (part.p0(1))> (E*part.w0(1)-part.q0(1)) 
        y=[-1;0];
    else
        y=[0;1];
    end
  else
   ma=max(abs(part.bcs(1)),abs(part.bcs(2)));
   y=[-part.bcs(2)/ma;part.bcs(1)/ma]; 
  end
  %shoot from a to b
  y=propagate_solution(part,E,y,imatch,true);
  phi=y(1);
else
  if part.inda>0
    if E*part.w0(1)-part.q0(1)<0
        yi=[0;1];
    elseif (part.p0(1))> (E*part.w0(1)-part.q0(1)) 
        yi=[-1;0];
    else
        yi=[0;1];
    end
  else
    ma=max(abs(part.bcs(1)),abs(part.bcs(2)));
    yi=[-part.bcs(2)/ma;part.bcs(1)/ma];
  end
  %better from the two ends to the middle!
  ma=max(part.inda+1,part.inf_a+1);
  mi=length(part.w0)-max(part.indb,part.inf_b);
  [m,imatch]=max((E*part.w0(ma:mi)-part.q0(ma:mi)).*part.p0(ma:mi));
  imatch=imatch+ma-1;
  %shoot from a to matching point
  yl = propagate_solution(part,E,yi,imatch,true);
  if part.indb>0
    if E*part.w0(end)-part.q0(end)<0
        yi=[0;1];
    elseif (part.p0(end))> (E*part.w0(end)-part.q0(end)) 
        yi=[-1;0];
    else
        yi=[0;1];
    end
  else
    ma=max(abs(part.bcs(3)),abs(part.bcs(4)));
    yi=[-part.bcs(4)/ma;part.bcs(3)/ma];
  end
  %shoot from b to matching point
  yr = propagate_solution(part,E,yi,imatch,false);
  %mismatch:
  phi=yl(1)*yr(2)-yr(1)*yl(2);  
end
end


function [rp]=calculate_partition_refine_s(part,pfun,qfun,wfun)
%refines partition (mesh) in part
inda=0; indb=0;
if part.inf_a
    inda=1;
end
if part.inf_b
    indb=1;
end
xs=part.xs; xleft=xs(inda+1);
q0(1:inda)=part.q0(1:inda); p0(1:inda)=part.p0(1:inda); w0(1:inda)=part.w0(1:inda);
q1(1:inda)=part.q1(1:inda); p1(1:inda)=part.p1(1:inda); w1(1:inda)=part.w1(1:inda);
q2(1:inda)=part.q2(1:inda); p2(1:inda)=part.p2(1:inda); w2(1:inda)=part.w2(1:inda);
xsr(1:inda+1)=xs(1:inda+1);
hs=zeros(1,(length(xs)-1-indb-inda)*2);
xst=zeros(1,(length(xs)-1-indb-inda)*2);
xsrt=zeros(1,(length(xs)-1-indb-inda)*2);
j=1;
for i=inda+1:length(xs)-1-indb
    h=xs(i+1)-xleft;
    h2=h/2;
    xst(j:j+1)=[xleft xleft+h2];
    hs(j:j+1)=[h2 h2]; 
    xsrt(j:j+1)=[xleft+h2 xleft+2*h2];
    j=j+2;
    xleft=xleft+h;
end
xsr=[xsr xsrt];
[q0t,q1t,q2t]=get_LegendrePolyn3(qfun,xst,hs);
[w0t,w1t,w2t]=get_LegendrePolyn3(wfun,xst,hs);
[p0t,p1t,p2t]=get_LegendrePolyn4(pfun,xst,hs);
q0=[q0 q0t]; q1=[q1 q1t]; q2=[q2 q2t];
w0=[w0 w0t]; w1=[w1 w1t]; w2=[w2 w2t];
p0=[p0 p0t]; p1=[p1 p1t]; p2=[p2 p2t];
mi=length(part.p0)-indb+1;
rp.p0=[p0 part.p0(mi:end)];
rp.q0=[q0 part.q0(mi:end)];
rp.w0=[w0 part.w0(mi:end)];
rp.p1=[p1 part.p1(mi:end)];
rp.q1=[q1 part.q1(mi:end)];
rp.w1=[w1 part.w1(mi:end)];
rp.p2=[p2 part.p2(mi:end)];
rp.q2=[q2 part.q2(mi:end)];
rp.w2=[w2 part.w2(mi:end)];
rp.xs=[xsr xs(end-indb+1:end)];
rp.inf_b=part.inf_b;
rp.inf_a=part.inf_a;
rp.bcs=part.bcs;
rp.inda=part.inda;
rp.indb=part.indb;
end
 

function [rp]=calculate_partition_refine(part,pfun,qfun,wfun)
%refines mesh in part
inda=part.inda; indb=part.indb;
if part.inf_a
    inda=1;
end
if part.inf_b
    indb=1;
end
xs=part.xs; xleft=xs(inda+1);
q0(1:inda)=part.q0(1:inda); p0(1:inda)=part.p0(1:inda); w0(1:inda)=part.w0(1:inda);
q1(1:inda)=part.q1(1:inda); p1(1:inda)=part.p1(1:inda); w1(1:inda)=part.w1(1:inda);
q2(1:inda)=part.q2(1:inda); p2(1:inda)=part.p2(1:inda); w2(1:inda)=part.w2(1:inda);
xsr(1:inda+1)=xs(1:inda+1);
hs=zeros(1,(length(xs)-1-indb-inda)*2);
xst=zeros(1,(length(xs)-1-indb-inda)*2);
xsrt=zeros(1,(length(xs)-1-indb-inda)*2);
j=1;
for i=inda+1:length(xs)-1-indb
    h=xs(i+1)-xleft;
    h2=h/2;
    xst(j:j+1)=[xleft xleft+h2];
    hs(j:j+1)=[h2 h2]; 
    xsrt(j:j+1)=[xleft+h2 xleft+2*h2];
    j=j+2;
    xleft=xleft+h;
end
xsr=[xsr xsrt];
[q0t,q1t,q2t]=get_LegendrePolyn3(qfun,xst,hs);
[w0t,w1t,w2t]=get_LegendrePolyn3(wfun,xst,hs);
[p0t,p1t,p2t]=get_LegendrePolyn4(pfun,xst,hs);
q0=[q0 q0t]; q1=[q1 q1t]; q2=[q2 q2t];
w0=[w0 w0t]; w1=[w1 w1t]; w2=[w2 w2t];
p0=[p0 p0t]; p1=[p1 p1t]; p2=[p2 p2t];

mi=length(part.p0)-indb+1;
rp.p0=[p0 part.p0(mi:end)];
rp.q0=[q0 part.q0(mi:end)];
rp.w0=[w0 part.w0(mi:end)];
rp.p1=[p1 part.p1(mi:end)];
rp.q1=[q1 part.q1(mi:end)];
rp.w1=[w1 part.w1(mi:end)];
rp.p2=[p2 part.p2(mi:end)];
rp.q2=[q2 part.q2(mi:end)];
rp.w2=[w2 part.w2(mi:end)];
rp.xs=[xsr xs(end-indb+1:end)];
rp.inf_b=part.inf_b;
rp.inf_a=part.inf_a;
rp.bcs=part.bcs;
rp.inda=part.inda;
rp.indb=part.indb;
end
%  

function [rp]=calculate_partition_refine_sing(part,pfun,qfun,wfun)
%refines mesh near singular endpoint
p0=[]; q0=[]; w0=[]; p1=[]; q1=[]; w1=[]; p2=[]; q2=[]; w2=[]; xs=[];
inda=part.inda; indb=part.indb;
if inda>1 %&& part.indb==0
 N=inda*2;
 p0=zeros(1,N); q0=zeros(1,N); w0=zeros(1,N); p1=zeros(1,N); q1=zeros(1,N); 
 w1=zeros(1,N); p2=zeros(1,N); q2=zeros(1,N); w2=zeros(1,N); 
 hs=part.xs(2:inda+1)-part.xs(1:inda);
 hx=sum(hs)/((N+1)*N/2);
 x=part.xs(1);
 xs=x;
 for i=1:N
     h=i*hx; 
     [q0(i),q1(i),q2(i)]=get_LegendrePolyn(qfun,x,h);
     [w0(i),w1(i),w2(i)]=get_LegendrePolyn(wfun,x,h);
     [p0(i),p1(i),p2(i)]=get_LegendrePolyn2(pfun,x,h);
     x=x+h;
     xs=[xs x];
 end
 xs=xs(1:end-1);
end
bp0=[]; bq0=[]; bw0=[]; bp1=[]; bq1=[]; bw1=[]; bp2=[]; bq2=[]; bw2=[];
xs2=[];
if indb>1 
 N=indb*2;   
 bp0=zeros(1,N); bq0=zeros(1,N); bw0=zeros(1,N); bp1=zeros(1,N); 
 bq1=zeros(1,N); bw1=zeros(1,N); bp2=zeros(1,N); bq2=zeros(1,N); 
 bw2=zeros(1,N);
 hs=part.xs(end-indb+1:end)-part.xs(end-indb:end-1);  
 hx=sum(hs)/((N+1)*N/2);
 x=part.xs(end-indb);
 xs2=x;
 for i=1:N
     h=i*hx; 
     [bq0(i),bq1(i),bq2(i)]=get_LegendrePolyn(qfun,x,h);
     [bw0(i),bw1(i),bw2(i)]=get_LegendrePolyn(wfun,x,h);
     [bp0(i),bp1(i),bp2(i)]=get_LegendrePolyn2(pfun,x,h);
     x=x+h;
     xs2=[xs2 x];
 end
end
rp.xs=[xs part.xs(inda+1:end-indb) xs2(2:end)];
rp.p0=[p0 part.p0(inda+1:end-indb) bp0];
rp.q0=[q0 part.q0(inda+1:end-indb) bq0];
rp.w0=[w0 part.w0(inda+1:end-indb) bw0];
rp.p1=[p1 part.p1(inda+1:end-indb) bp1];
rp.q1=[q1 part.q1(inda+1:end-indb) bq1];
rp.w1=[w1 part.w1(inda+1:end-indb) bw1];
rp.p2=[p2 part.p2(inda+1:end-indb) bp2];
rp.q2=[q2 part.q2(inda+1:end-indb) bq2];
rp.w2=[w2 part.w2(inda+1:end-indb) bw2];
rp.inda=inda*2;
rp.indb=indb*2;
rp.inf_b=part.inf_b;
rp.inf_a=part.inf_a;
rp.bcs=part.bcs;
end




function [V0,V1,V2]=get_LegendrePolyn(fun,x,h)
%Gauss Legendre (3 Legendre points)
cc=sqrt(15)/5;
c=h/2*[-cc+1 1 cc+1];
f=fun(x+c);
f=f.*ones(1,3); %to avoid problems when e.g. fun=1 
V0=1/18*(8*f(2)+5*f(1)+5*f(3));
V1=(3*(-1/18*15^(1/2)*(f(1)-f(3))))*h^2;
V2=5/9*(-2*f(2)+f(1)+f(3))*h^2;
end

function [V0,V1,V2]=get_LegendrePolyn2(fun,x,h)
cc=sqrt(15)/5;
c=h/2*[-cc+1 1 cc+1];
f=1./fun(x+c);
f=f.*ones(1,3); %to avoid problems when e.g. fun=1 
V0=1/18*(8*f(2)+5*f(1)+5*f(3));
V1=(3*(-1/18*15^(1/2)*(f(1)-f(3))))*h^2;
V2=5/9*(-2*f(2)+f(1)+f(3))*h^2;
end

function [V0,V1,V2]=get_LegendrePolyn3(fun,x,h)
%faster version
%Gauss Legendre (3 Legendre points)
xs=zeros(1,length(x)*3);
cc=sqrt(15)/5;
len=length(x);
for i=1:len
 c=h(i)/2*[-cc+1 1 cc+1];
 xs((i-1)*3+1:i*3)=x(i)+c;
end
fs=fun(xs);
fs=fs.*ones(1,length(xs)); %to avoid problems when e.g. fun=1 
V0=zeros(1,len);
V1=zeros(1,len);
V2=zeros(1,len);
for i=1:len
 f=fs((i-1)*3+1:i*3);
 V0(i)=1/18*(8*f(2)+5*f(1)+5*f(3));
 V1(i)=(3*(-1/18*15^(1/2)*(f(1)-f(3))))*h(i)^2;
 V2(i)=5/9*(-2*f(2)+f(1)+f(3))*h(i)^2;
end
end

function [V0,V1,V2]=get_LegendrePolyn4(fun,x,h)
%faster version of get_LegendrePolyn2
xs=zeros(1,length(x)*3);
cc=sqrt(15)/5;
len=length(x);
for i=1:len
 c=h(i)/2*[-cc+1 1 cc+1];
 xs((i-1)*3+1:i*3)=x(i)+c;
end
fs=1./fun(xs);
fs=fs.*ones(1,length(xs)); %to avoid problems when e.g. fun=1 
V0=zeros(1,len);
V1=zeros(1,len);
V2=zeros(1,len);
for i=1:len
 f=fs((i-1)*3+1:i*3);
 V0(i)=1/18*(8*f(2)+5*f(1)+5*f(3));
 V1(i)=(3*(-1/18*15^(1/2)*(f(1)-f(3))))*h(i)^2;
 V2(i)=5/9*(-2*f(2)+f(1)+f(3))*h(i)^2;
end
end




